import time

from dagster import Config, Definitions, OpExecutionContext, job, op


class MyConfig(Config):
    sleep_for: int = 3


@op
def return_one(context: OpExecutionContext, config: MyConfig) -> int:
    context.log.debug(f"sleeping for {config.sleep_for} seconds...")
    time.sleep(config.sleep_for)

    return 1


@op
def return_two(context: OpExecutionContext, config: MyConfig) -> int:
    context.log.info(f"sleeping for {config.sleep_for} seconds...")
    time.sleep(config.sleep_for)
    context.log.info("Waking up!")

    return 2


@op
def sum_one_and_two(a: int, b: int) -> int:
    res = a + b

    assert res == 3

    return res


@job(tags={"dagster-ray/config": {"num_cpus": 0.5}})
def my_job():
    return_two_result = return_two()
    return_one_result = return_one()

    sum_one_and_two(return_one_result, return_two_result)


definitions = Definitions(
    jobs=[my_job],
)



import os

DEFAULT_DEPLOYMENT_NAME = (
    os.getenv("DAGSTER_CLOUD_DEPLOYMENT_NAME")
    if os.getenv("DAGSTER_CLOUD_IS_BRANCH_DEPLOYMENT") == "0"
    else os.getenv("DAGSTER_CLOUD_GIT_BRANCH")
) or "dev"

IS_PROD = DEFAULT_DEPLOYMENT_NAME == "prod"



from __future__ import annotations

import logging
import socket
import subprocess
import threading
import time
from collections.abc import Iterator
from contextlib import contextmanager
from io import FileIO
from queue import Queue
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    TypedDict,
    cast,
)

from typing_extensions import NotRequired

from dagster_ray.kuberay.client.base import BaseKubeRayClient

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from kubernetes.client import ApiClient
    from ray.job_submission import JobSubmissionClient


GROUP = "ray.io"
VERSION = "v1"
PLURAL = "rayclusters"
KIND = "RayCluster"


def enqueue_output(out: FileIO, queue: Queue, should_stop: Callable[[], bool]):
    import select

    try:
        while True:
            if should_stop():
                break
            rlist, _, _ = select.select([out], [], [], 1.0)
            if not rlist:
                continue
            line = out.readline()
            if not line:
                break
            queue.put(line)
    except Exception:
        logger.exception("Exception encountered while reading port-forwarded RayCluster logs")
    finally:
        try:
            out.close()
        except Exception:
            logger.exception("Exception encountered while closing RayCluster logs during port-forwarding")


def get_random_available_port() -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))  # Bind to an available port chosen by the OS
        _, port = s.getsockname()  # Get the port number assigned
        return port


class RayClusterEndpoints(TypedDict):  # these are ports
    client: str
    dashboard: str
    metrics: str
    redis: str
    serve: str


class RayClusterHead(TypedDict):
    podIP: str
    serviceIP: str


class RayClusterStatus(TypedDict):
    desiredCPU: str
    desiredGPU: str
    desiredMemory: str
    desiredTPU: str

    lastUpdateTime: str  # "2024-09-16T09:34:29Z"
    maxWorkerReplicas: int
    observedGeneration: int

    head: NotRequired[RayClusterHead]
    endpoints: NotRequired[RayClusterEndpoints]
    state: NotRequired[str]


class RayClusterClient(BaseKubeRayClient[RayClusterStatus]):
    def __init__(
        self,
        config_file: str | None = None,
        context: str | None = None,
        api_client: ApiClient | None = None,
    ) -> None:
        super().__init__(group=GROUP, version=VERSION, kind=KIND, plural=PLURAL, api_client=api_client)

        # these are only used because of kubectl port-forward CLI command
        # TODO: remove kubectl usage and remove these attributes
        self.config_file = config_file
        self.context = context

    def wait_until_ready(
        self,
        name: str,
        namespace: str,
        timeout: int,
        image: str | None = None,
        log_cluster_conditions: bool = False,
    ) -> tuple[str, RayClusterEndpoints]:
        from kubernetes import watch

        """
        If ready, returns service ip address and a dictionary of ports.
        """

        w = watch.Watch()

        start_time = time.time()

        # TODO: use get_namespaced_custom_object instead
        # once https://github.com/kubernetes-client/python/issues/1679
        # is solved

        condition_index_to_log = 0

        for event in w.stream(
            self._api.list_namespaced_custom_object,
            self.group,
            self.version,
            namespace,
            self.plural,
        ):
            item = cast(dict[str, Any], event["raw_object"])  # type: ignore

            if "status" not in item:
                continue

            status: RayClusterStatus = item["status"]
            state = status.get("state")

            # https://docs.ray.io/en/latest/cluster/kubernetes/user-guides/observability.html#raycluster-status-conditions
            conditions = list(reversed(status.get("conditions", [])))

            if log_cluster_conditions:
                while len(conditions) > condition_index_to_log:
                    logger.info(f"RayCluster {namespace}/{name} condition: {conditions[condition_index_to_log]}")
                    condition_index_to_log += 1

            if state == "failed":
                raise Exception(
                    f"RayCluster {namespace}/{name} failed to start. Status:\n\n{status}\n\nMore details: `kubectl -n {namespace} describe RayCluster {name}`"
                )

            if (
                item.get("metadata")  # type: ignore
                and item["metadata"].get("name") == name  # type: ignore
                and state == "ready"
                and status.get("head")
                and status.get("endpoints", {}).get("dashboard")
            ):
                if image is not None:
                    if (
                        item.get("spec")  # type: ignore
                        and item["spec"]["headGroupSpec"]["template"]["spec"]["containers"][0]["image"]  # type: ignore
                        != image
                    ):
                        continue
                w.stop()
                logger.debug(f"RayCluster {namespace}/{name} is ready!")
                return status["head"]["serviceIP"], status["endpoints"]

            if time.time() - start_time > timeout:
                w.stop()
                raise TimeoutError(f"Timed out waiting for RayCluster {namespace}/{name} to be ready. Status: {status}")

        raise Exception("This code should be unreachable")

    @contextmanager
    def port_forward(
        self,
        name: str,
        namespace: str,
        local_dashboard_port: int = 8265,
        local_gcs_port: int = 10001,
    ) -> Iterator[tuple[int, int]]:
        """
        Port forwards the Ray dashboard and GCS ports to localhost.
        Use 0 for local_dashboard_port and local_gcs_port to get random available ports.
        Returns the ports that the dashboard and GCS are forwarded to.
        """
        if local_dashboard_port == 0:
            local_dashboard_port = get_random_available_port()

        if local_gcs_port == 0:
            local_gcs_port = get_random_available_port()

        service = f"{name}-head-svc"

        if not self.get_status(name, namespace):
            raise RuntimeError(f"RayCluster {name} does not exist in namespace {namespace}")

        self.wait_for_service_endpoints(service_name=service, namespace=namespace)

        cmd = [
            "kubectl",
            "port-forward",
            "-n",
            namespace,
            f"svc/{service}",
            f"{local_dashboard_port}:8265",
            f"{local_gcs_port}:10001",
        ]

        if self.context:
            cmd.extend(["--context", self.context])

        if self.config_file:
            cmd.extend(["--kubeconfig", self.config_file])

        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

        queue = Queue()
        SHOULD_STOP = False

        def should_stop():
            return SHOULD_STOP

        t = threading.Thread(target=enqueue_output, args=(process.stdout, queue, should_stop))
        t.daemon = True  # thread dies with the program
        t.start()

        time.sleep(0.01)

        try:
            # inspect stdout to avoid connecting too early
            # we must check for "Forwarding from" in stdout
            #

            while True:
                if process.poll() is not None:
                    raise RuntimeError(
                        f"port-forwarding command: `{' '.join(cmd)}` failed. Most likely ports {local_dashboard_port} or {local_gcs_port} are  already in use, or Kubernetes service {name}-head-svc does not exist in namespace {namespace}."
                    )

                line = queue.get()

                if "Forwarding from" in line:
                    break

            logger.info(
                f"Connecting to {namespace}/{name} via port-forwarding for ports {local_dashboard_port} and {local_gcs_port}..."
            )

            yield local_dashboard_port, local_gcs_port
        finally:
            # terminate the thread
            SHOULD_STOP = True
            process.kill()
            process.wait()
            t.join()
            logger.info(f"Port-forwarding for ports {local_dashboard_port} and {local_gcs_port} has been stopped.")

    @contextmanager
    def job_submission_client(
        self, name: str, namespace: str, port_forward: bool = False, timeout: int = 60
    ) -> Iterator[JobSubmissionClient]:
        """
        Returns a JobSubmissionClient object that can be used to interact with Ray jobs running in the KubeRay cluster.
        If port_forward is True, it will port forward the dashboard and GCS ports to localhost, and should be used in a context manager.
        If port_forward is False, the client will connect to the dashboard directly (assuming the dashboard is accessible from the host).
        """

        from ray.job_submission import JobSubmissionClient

        if not port_forward:
            # TODO: revisit the decision to use this as context-manager in this case
            #
            status = self.get_status(name, namespace)

            host = status["head"]["serviceIP"]  # type: ignore
            dashboard_port = status["endpoints"]["dashboard"]  # type: ignore

            yield JobSubmissionClient(address=f"http://{host}:{dashboard_port}")
        else:
            self.wait_for_service_endpoints(service_name=f"{name}-head-svc", namespace=namespace, timeout=timeout)
            with self.port_forward(name=name, namespace=namespace, local_dashboard_port=0, local_gcs_port=0) as (
                local_dashboard_port,
                _,
            ):
                yield JobSubmissionClient(address=f"http://localhost:{local_dashboard_port}")



from dagster import ScheduleDefinition

from dagster_ray.kuberay.jobs import cleanup_kuberay_clusters

cleanup_kuberay_clusters_daily = ScheduleDefinition(
    job=cleanup_kuberay_clusters,
    cron_schedule="0 0 * * *",
    name="cleanup_kuberay_clusters_schedule_daily",
)



from __future__ import annotations

import contextlib
import hashlib
import random
import re
import string
import sys
from collections.abc import Generator
from typing import TYPE_CHECKING, Any, cast

import dagster._check as check
from dagster import ConfigurableResource, InitResourceContext
from dagster._annotations import beta
from pydantic import Field, PrivateAttr

from dagster_ray._base.constants import DEFAULT_DEPLOYMENT_NAME
from dagster_ray.kuberay.client import RayClusterClient

# yes, `python-client` is actually the KubeRay package name
# https://github.com/ray-project/kuberay/issues/2078
from dagster_ray.kuberay.configs import RayClusterConfig
from dagster_ray.kuberay.utils import normalize_k8s_label_values

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from dagster import DagsterRun, DagsterRunStatus
from ray._private.worker import BaseContext as RayBaseContext  # noqa

from dagster_ray._base.resources import BaseRayResource
from dagster_ray.kuberay.client.base import load_kubeconfig

if TYPE_CHECKING:
    import kubernetes


@beta
class RayClusterClientResource(ConfigurableResource):
    kube_context: str | None = None
    kubeconfig_file: str | None = None

    _raycluster_client: RayClusterClient = PrivateAttr()
    _k8s_api: kubernetes.client.CustomObjectsApi = PrivateAttr()
    _k8s_core_api: kubernetes.client.CoreV1Api = PrivateAttr()

    @property
    def client(self) -> RayClusterClient:
        if self._raycluster_client is None:
            raise ValueError(f"{self.__class__.__name__} not initialized")
        return self._raycluster_client

    @property
    def k8s(self) -> kubernetes.client.CustomObjectsApi:
        if self._k8s_api is None:
            raise ValueError(f"{self.__class__.__name__} not initialized")
        return self._k8s_api

    @property
    def k8s_core(self) -> kubernetes.client.CoreV1Api:
        if self._k8s_core_api is None:
            raise ValueError(f"{self.__class__.__name__} not initialized")
        return self._k8s_core_api

    def setup_for_execution(self, context: InitResourceContext) -> None:
        import kubernetes

        load_kubeconfig(context=self.kube_context, config_file=self.kubeconfig_file)

        self._raycluster_client = RayClusterClient(context=self.kube_context, config_file=self.kubeconfig_file)
        self._k8s_api = kubernetes.client.CustomObjectsApi()
        self._k8s_core_api = kubernetes.client.CoreV1Api()


@beta
class KubeRayCluster(BaseRayResource):
    """
    Provides a RayCluster for the current step selection
    The cluster is automatically deleted after steps execution
    """

    deployment_name: str = Field(
        default=DEFAULT_DEPLOYMENT_NAME,
        description="Prefix for the RayCluster name. Dagster Cloud variables are used to determine the default value.",
    )
    env_vars: list[Any] = Field(
        default_factory=list,
        description="Environment variables to inject into all RayCluster containers in Kubernetes format.",
    )
    ray_cluster: RayClusterConfig = Field(default_factory=RayClusterConfig)
    skip_cleanup: bool = False
    skip_init: bool = False
    timeout: int = Field(default=600, description="Timeout in seconds for the RayCluster to become ready")
    log_cluster_conditions: bool = Field(
        default=True,
        description="Whether to log RayCluster conditions while waiting for the RayCluster to become ready. For more information, see https://docs.ray.io/en/latest/cluster/kubernetes/user-guides/observability.html#raycluster-status-conditions.",
    )

    client: RayClusterClientResource = Field(default_factory=RayClusterClientResource)

    _cluster_name: str = PrivateAttr()
    _host: str = PrivateAttr()

    @property
    def host(self) -> str:
        if self._host is None:
            raise ValueError(f"{self.__class__.__name__} not initialized")
        return self._host

    @property
    def namespace(self) -> str:
        return self.ray_cluster.namespace  # TODO: add namespace auto-creation logic

    @property
    def cluster_name(self) -> str:
        if self._cluster_name is None:
            raise ValueError(f"{self.__class__.__name__}not initialized")
        return self._cluster_name

    def get_dagster_tags(self, context: InitResourceContext) -> dict[str, str]:
        tags = super().get_dagster_tags(context=context)
        tags.update({"dagster.io/cluster": self.cluster_name, "dagster.io/deployment": self.deployment_name})
        return tags

    @contextlib.contextmanager
    def yield_for_execution(self, context: InitResourceContext) -> Generator[Self, None, None]:
        assert context.log is not None
        assert context.dagster_run is not None

        self.client.setup_for_execution(context)

        self._cluster_name = self._get_ray_cluster_step_name(context)

        try:
            # just a safety measure, no need to recreate the cluster for step retries or smth
            if not self.client.client.list(
                namespace=self.namespace,
                label_selector=f"dagster.io/cluster={self.cluster_name}",
            )["items"]:
                cluster_body = self._build_raycluster(
                    context,
                    image=(self.ray_cluster.image or context.dagster_run.tags["dagster/image"]),
                    labels=normalize_k8s_label_values(self.get_dagster_tags(context)),
                )

                resource = self.client.client.create(body=cluster_body, namespace=self.namespace)
                if not resource:
                    raise RuntimeError(f"Couldn't create RayCluster {self.namespace}/{self.cluster_name}")

                context.log.info(
                    f"Created RayCluster {self.namespace}/{self.cluster_name}. Waiting for it to become ready..."
                )

                self._wait_raycluster_ready()

                self._host = self.client.client.get_status(name=self.cluster_name, namespace=self.namespace)[  # type: ignore
                    "head"
                ]["serviceIP"]

                msg = f"RayCluster {self.namespace}/{self.cluster_name} is ready! Connection command:\n"
                msg += f"kubectl -n {self.namespace} port-forward svc/{self.cluster_name}-head-svc 8265:8265 6379:6379 10001:10001"

                context.log.info(msg)

            if not self.skip_init:
                self.init_ray(context)
            else:
                self._context = None

            yield self
        except Exception as e:
            context.log.critical(f"Couldn't create or connect to RayCluster {self.namespace}/{self.cluster_name}!")
            self._maybe_cleanup_raycluster(context)
            raise e

        self._maybe_cleanup_raycluster(context)
        if self._context is not None:
            self._context.disconnect()

    def _build_raycluster(
        self,
        context: InitResourceContext,
        image: str | None = None,
        labels: dict[str, str] | None = None,  # TODO: use in RayCluster labels
    ) -> dict[str, Any]:
        assert context.log is not None
        """
        Builds a RayCluster from the provided configuration, while injecting custom image and labels (only known during resource setup)
        """

        # TODO: inject self.redis_port and self.dashboard_port into the RayCluster configuration
        # TODO: autoa-apply some tags from dagster-k8s/config

        labels = labels or {}
        assert isinstance(labels, dict)

        image = self.ray_cluster.image or image
        head_group_spec = self.ray_cluster.head_group_spec.copy()
        worker_group_specs = self.ray_cluster.worker_group_specs.copy()

        env_vars = self.env_vars.copy()

        env_vars_to_inject = self.get_env_vars_to_inject()

        if env_vars_to_inject:
            context.log.info("Injecting debugging environment variables into the RayCluster configuration")
            for key, value in self.get_env_vars_to_inject().items():
                env_vars.append({"name": key, "value": value})

        def update_group_spec(group_spec: dict[str, Any]):
            # TODO: only inject if the container has a `dagster.io/inject-image` annotation or smth
            if group_spec["template"]["spec"]["containers"][0].get("image") is None:
                group_spec["template"]["spec"]["containers"][0]["image"] = image

            for container in group_spec["template"]["spec"]["containers"]:
                container["env"] = container.get("env", []) + env_vars

            if group_spec.get("metadata") is None:
                group_spec["metadata"] = {"labels": labels}
            else:
                if group_spec["metadata"].get("labels") is None:
                    group_spec["metadata"]["labels"] = labels
                else:
                    group_spec["metadata"]["labels"].update(labels)

        update_group_spec(head_group_spec)
        for worker_group_spec in worker_group_specs:
            update_group_spec(worker_group_spec)

        return {
            "apiVersion": "ray.io/v1",
            "kind": "RayCluster",
            "metadata": {
                "name": self.cluster_name,
                "labels": labels or {},
            },
            "spec": {
                "enableInTreeAutoscaling": self.ray_cluster.enable_in_tree_autoscaling,
                "autoscalerOptions": self.ray_cluster.autoscaler_options,
                "headGroupSpec": head_group_spec,
                "workerGroupSpecs": worker_group_specs,
            },
        }

    def _wait_raycluster_ready(self):
        import kubernetes

        self.client.client.wait_until_ready(
            self.cluster_name,
            namespace=self.namespace,
            timeout=self.timeout,
            log_cluster_conditions=self.log_cluster_conditions,
        )

        # the above code only checks for RayCluster creation
        # not for head pod readiness

        w = kubernetes.watch.Watch()
        for event in w.stream(
            func=self.client.k8s_core.list_namespaced_pod,
            namespace=self.namespace,
            label_selector=f"ray.io/cluster={self.cluster_name},ray.io/group=headgroup",
            timeout_seconds=60,
        ):
            if event["object"].status.phase == "Running":  # type: ignore
                w.stop()
                return
            # event.type: ADDED, MODIFIED, DELETED
            if event["type"] == "DELETED":  # type: ignore
                # Pod was deleted while we were waiting for it to start.
                w.stop()
                return

    def _maybe_cleanup_raycluster(self, context: InitResourceContext):
        assert context.log is not None

        if not self.skip_cleanup and cast(DagsterRun, context.dagster_run).status != DagsterRunStatus.FAILURE:
            self.client.client.delete(self.cluster_name, namespace=self.namespace)
            context.log.info(f"Deleted RayCluster {self.namespace}/{self.cluster_name}")
        else:
            context.log.warning(
                f"Skipping RayCluster {self.cluster_name} deletion because `disable_cluster_cleanup` "
                f"config parameter is set to `True` and the run has failed."
            )

    def _get_ray_cluster_step_name(self, context: InitResourceContext) -> str:
        assert isinstance(context.run_id, str)
        assert context.dagster_run is not None

        # try to make the name as short as possible
        cluster_name_prefix = f"dg-{self.deployment_name.replace('-', '')[:8]}-{context.run_id[:8]}"

        dagster_user_email = context.dagster_run.tags.get("user")
        if dagster_user_email is not None:
            cluster_name_prefix += f"-{dagster_user_email.replace('.', '').replace('-', '').split('@')[0][:6]}"

        step_key = self._get_step_key(context)

        name_key = get_k8s_object_name(
            context.run_id,
            step_key,
        )

        step_name = f"{cluster_name_prefix}-{name_key}".lower()
        step_name = re.sub(r"[^-0-9a-z]", "-", step_name)

        return step_name


def get_k8s_object_name(run_id: str, step_key: str | None = None):
    """Creates a unique (short!) identifier to name k8s objects based on run ID and step key(s).

    K8s Job names are limited to 63 characters, because they are used as labels. For more info, see:

    https://kubernetes.io/docs/concepts/overview/working-with-objects/names/
    """
    check.str_param(run_id, "run_id")
    check.opt_str_param(step_key, "step_key")
    if not step_key:
        letters = string.ascii_lowercase
        step_key = "".join(random.choice(letters) for i in range(20))

    # Creates 32-bit signed int, so could be negative
    name_hash = hashlib.md5((run_id + step_key).encode("utf-8"))

    return name_hash.hexdigest()[:8]



from dagster_ray.kuberay.utils import normalize_k8s_label_values


def test_normalize_k8s_label_values():
    assert normalize_k8s_label_values(
        {
            "foo": "bar",
            "my/label": "my/value",
            "user": "daniel@my.org",
            "user-dirty": "daniel!`~@my.org",
            "alphanumeric": "abc123",
            "long": 64 * "a",
            "badstart": "-foo",
            "badstart_after_initial_replace": "@foo",
        }
    ) == {
        "foo": "bar",
        "my/label": "myvalue",
        "user": "daniel-my-org",
        "user-dirty": "daniel-my-org",
        "alphanumeric": "abc123",
        "long": 63 * "a",
        "badstart": "foo",
        "badstart_after_initial_replace": "foo",
    }



from __future__ import annotations

import asyncio
import logging
import random
import string
import threading
import time
from collections.abc import AsyncIterator, Generator, Iterator
from contextlib import contextmanager
from functools import partial
from typing import TYPE_CHECKING, Any, TypedDict, Union, cast

import dagster._check as check
from dagster import AssetExecutionContext, OpExecutionContext, PipesClient
from dagster._annotations import beta
from dagster._core.definitions.resource_annotation import TreatAsResourceParam
from dagster._core.errors import DagsterExecutionInterruptedError
from dagster._core.pipes.client import (
    PipesClientCompletedInvocation,
    PipesContextInjector,
    PipesMessageReader,
)
from dagster._core.pipes.context import PipesLaunchedData, PipesMessageHandler, PipesSession
from dagster._core.pipes.utils import (
    PipesEnvContextInjector,
    _join_thread,
    extract_message_or_forward_to_stdout,
    open_pipes_session,
)
from dagster_pipes import PipesDefaultMessageWriter, PipesExtras, PipesParams
from typing_extensions import NotRequired, TypeAlias

from dagster_ray._base.utils import get_dagster_tags

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ray.job_submission import JobStatus, JobSubmissionClient


OpOrAssetExecutionContext: TypeAlias = Union[OpExecutionContext, AssetExecutionContext]


PIPES_LAUNCHED_EXTRAS_RAY_ADDRESS_KEY = "ray_address"
PIPES_LAUNCHED_EXTRAS_RAY_JOB_ID_KEY = "ray_job_id"


def tail_logs(tailer: AsyncIterator[str]) -> Generator[str | None, None, None]:
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    while True:
        try:
            yield loop.run_until_complete(tailer.__anext__())  # type: ignore
        except (StopAsyncIteration, ValueError):
            break
        except Exception:
            logger.exception("Encountered exception during ray job log tailing")
            yield None


@beta
class PipesRayJobMessageReader(PipesMessageReader):
    f"""
    Dagster Pipes message reader for receiving messages from a Ray job.
    Will extract Dagster events and forward the rest to stdout.

    Waits for PipesSession.report_launched to be called with `{PIPES_LAUNCHED_EXTRAS_RAY_ADDRESS_KEY}` and `{PIPES_LAUNCHED_EXTRAS_RAY_JOB_ID_KEY}` keys
    populated in `extras` in order to start message reading.

    Args:
        _job_submission_client_kwargs (Optional[dict[str, Any]]): additional arguments for `ray.job_submission.JobSubmissionClient`. The `address` value is expected to be set via `PipesSession.report_launched`, while other arguments can be set via this parameter.
    """

    _client: JobSubmissionClient | None
    _job_id: str | None

    def __init__(self, job_submission_client_kwargs: dict[str, Any] | None = None):
        self._job_submission_client_kwargs = job_submission_client_kwargs
        self._thread: threading.Thread | None = None
        self.session_closed = threading.Event()
        self._job_id = None
        self._client = None
        self.thread_ready = threading.Event()

        self.completed = threading.Event()

    @property
    def client(self) -> JobSubmissionClient:
        if self._client is None:
            raise RuntimeError(
                "client is only available after it has been seeded in the Pipes client via PipesSession.report_launched"
            )
        else:
            return self._client

    @property
    def job_id(self) -> str:
        if self._job_id is None:
            raise RuntimeError(
                "job_id is only available after it has been seeded in the Pipes client via PipesSession.report_launched"
            )
        else:
            return self._job_id

    def on_launched(self, launched_payload: PipesLaunchedData) -> None:
        from ray.job_submission import JobSubmissionClient

        self._client = JobSubmissionClient(
            address=launched_payload["extras"][PIPES_LAUNCHED_EXTRAS_RAY_ADDRESS_KEY],
            **(self._job_submission_client_kwargs or {}),
        )
        self._job_id = launched_payload["extras"][PIPES_LAUNCHED_EXTRAS_RAY_JOB_ID_KEY]

        super().on_launched(launched_payload)

    def messages_are_readable(self, params: PipesParams) -> bool:
        return self._job_id is not None and self._client is not None

    @contextmanager
    def read_messages(self, handler: PipesMessageHandler) -> Iterator[PipesParams]:
        try:
            self._thread = threading.Thread(
                target=partial(self.handle_job_logs, handler),
                daemon=True,
            )
            self._thread.start()
            yield {PipesDefaultMessageWriter.STDIO_KEY: PipesDefaultMessageWriter.STDOUT}
        finally:
            self.terminate(handler)

    def terminate(self, handler: PipesMessageHandler) -> None:
        self.session_closed.set()

        if self._thread is not None:
            _join_thread(self._thread, "ray job logs tailing")
            if self._job_id is None:
                handler._context.log.warning(
                    "[Pipes] PipesRayJobMessageReader._job_id is not set. It's expected to be set via PipesSession.report_launched."
                )

    def no_messages_debug_text(self) -> str:
        return f"Tried to read messages from Ray Job logs, but no messages were received. Make sure PipesSession.report_launched has been called with `{PIPES_LAUNCHED_EXTRAS_RAY_JOB_ID_KEY}` set under `extras` field."

    def handle_job_logs(
        self,
        handler: PipesMessageHandler,
    ):
        try:
            while (self._job_id is None or self._client is None) and not self.session_closed.is_set():
                time.sleep(0.1)

            self.thread_ready.set()

            if self.session_closed.is_set():
                handler._context.log.warning(
                    "[pipes] PipesSession.report_launched was not called before session was closed."
                )
                return
            else:
                handler._context.log.debug("[pipes] PipesRayJobMessageReader is ready to begin reading ray job logs")

            assert self._job_id is not None and self._client is not None, (
                "Job ID and client are required for tailing logs. Expected on_launched to set them by this line."
            )

            session_closed_at = None

            tailer = self.client.tail_job_logs(job_id=self.job_id)

            for chunk in tail_logs(tailer):
                if chunk:
                    for log_line in chunk.split("\n"):
                        if log_line:
                            extract_message_or_forward_to_stdout(handler, log_line)

                if self.session_closed.is_set():
                    if session_closed_at is None:
                        session_closed_at = time.time()

                    if time.time() - session_closed_at > 30:  # wait for 30 seconds to flush all logs
                        self.completed.set()
                        return
        except:
            raise
        finally:
            self.completed.set()


class SubmitJobParams(TypedDict):
    entrypoint: str
    runtime_env: NotRequired[dict[str, Any]]
    metadata: NotRequired[dict[str, str]]
    submission_id: NotRequired[str]
    entrypoint_num_cpus: NotRequired[float]
    entrypoint_num_gpus: NotRequired[float]
    entrypoint_memory: NotRequired[int]
    entrypoint_resources: NotRequired[dict[str, float]]


class EnrichedSubmitJobParams(TypedDict):
    entrypoint: str
    runtime_env: dict[str, Any]
    metadata: dict[str, str]
    submission_id: str
    entrypoint_num_cpus: NotRequired[float]
    entrypoint_num_gpus: NotRequired[float]
    entrypoint_memory: NotRequired[int]
    entrypoint_resources: NotRequired[dict[str, float]]


def generate_job_id() -> str:
    letters = string.ascii_lowercase
    random_letters = "".join(random.choice(letters) for i in range(10))
    return f"pipes-{random_letters}"


@beta
class PipesRayJobClient(PipesClient, TreatAsResourceParam):
    """A pipes client for running Ray jobs on remote clusters.

    Starts the job directly on the Ray cluster and reads the logs from the job.

    Args:
        client (JobSubmissionClient): The Ray job submission client
        context_injector (Optional[PipesContextInjector]): A context injector to use to inject
            context into the Ray job. Defaults to :py:class:`PipesEnvContextInjector`.
        message_reader (Optional[PipesMessageReader]): A message reader to use to read messages
            from the glue job run. Defaults to :py:class:`PipesRayJobMessageReader`.
        forward_termination (bool): Whether to cancel the `RayJob` job run when the Dagster process receives a termination signal.
        timeout (int): Timeout for various internal interactions with the Kubernetes RayJob.
        poll_interval (int): Interval at which to poll the Kubernetes for status updates.
        Is useful when running in a local environment.
    """

    def __init__(
        self,
        client: JobSubmissionClient,
        context_injector: PipesContextInjector | None = None,
        message_reader: PipesMessageReader | None = None,
        forward_termination: bool = True,
        timeout: int = 600,
        poll_interval: int = 5,
    ):
        self.client = client
        self._context_injector = context_injector or PipesEnvContextInjector()
        self._message_reader = message_reader or PipesRayJobMessageReader()

        self.forward_termination = check.bool_param(forward_termination, "forward_termination")
        self.timeout = check.int_param(timeout, "timeout")
        self.poll_interval = check.int_param(poll_interval, "poll_interval")

        self._job_submission_client: JobSubmissionClient | None = None

    def run(  # type: ignore
        self,
        *,
        context: OpOrAssetExecutionContext,
        submit_job_params: SubmitJobParams,
        extras: PipesExtras | None = None,
    ) -> PipesClientCompletedInvocation:
        """
        Execute a RayJob, enriched with the Pipes protocol.

        Args:
            context (OpExecutionContext): Current Dagster op or asset context.
            ray_job (Dict[str, Any]): RayJob specification. `API reference <https://ray-project.github.io/kuberay/reference/api/#rayjob>`_.
            extras (Optional[Dict[str, Any]]): Additional information to pass to the Pipes session.
        """

        with open_pipes_session(
            context=context,
            message_reader=self._message_reader,
            context_injector=self._context_injector,
            extras=extras,
        ) as session:
            enriched_submit_job_params = self._enrich_submit_job_params(context, session, submit_job_params)

            job_id = self._start(context, session, enriched_submit_job_params)

            try:
                # self._read_messages(context, job_id)
                self._wait_for_completion(context, job_id)
                return PipesClientCompletedInvocation(session, metadata={"Ray Job ID": job_id})

            except DagsterExecutionInterruptedError:
                if self.forward_termination:
                    context.log.warning(f"[pipes] Dagster process interrupted! Will terminate RayJob {job_id}.")
                    self._terminate(context, job_id)
                raise

    def get_dagster_tags(self, context: OpOrAssetExecutionContext) -> dict[str, str]:
        tags = get_dagster_tags(context)
        return tags

    def _enrich_submit_job_params(
        self, context: OpOrAssetExecutionContext, session: PipesSession, submit_job_params: SubmitJobParams
    ) -> EnrichedSubmitJobParams:
        runtime_env = submit_job_params.get("runtime_env", {})
        metadata = submit_job_params.get("metadata", {})

        env_vars = session.get_bootstrap_env_vars()

        metadata.update(self.get_dagster_tags(context))
        runtime_env["env_vars"] = runtime_env.get("env_vars", {})
        runtime_env["env_vars"].update(env_vars)

        submit_job_params["metadata"] = metadata
        submit_job_params["runtime_env"] = runtime_env
        submit_job_params["submission_id"] = submit_job_params.get("submission_id", generate_job_id())

        return cast(EnrichedSubmitJobParams, submit_job_params)

    def _start(
        self, context: OpOrAssetExecutionContext, session: PipesSession, submit_job_params: EnrichedSubmitJobParams
    ) -> str:
        submission_id = submit_job_params["submission_id"]

        context.log.info(f"[pipes] Starting Ray job {submission_id}...")

        job_id = self.client.submit_job(**submit_job_params)

        context.log.info(f"[pipes] Ray job {job_id} started.")

        session.report_launched(
            {
                "extras": {
                    PIPES_LAUNCHED_EXTRAS_RAY_JOB_ID_KEY: job_id,
                    PIPES_LAUNCHED_EXTRAS_RAY_ADDRESS_KEY: self.client.get_address(),
                }
            }
        )

        return job_id

    def _wait_for_completion(self, context: OpOrAssetExecutionContext, job_id: str) -> JobStatus:
        from ray.job_submission import JobStatus

        context.log.info(f"[pipes] Waiting for RayJob {job_id} to complete...")

        while True:
            status = self.client.get_job_status(job_id)

            if status.is_terminal():
                if status in {JobStatus.FAILED, JobStatus.STOPPED}:
                    job_details = self.client.get_job_info(job_id)
                    raise RuntimeError(
                        f"[pipes] RayJob {job_id} failed with status {status}. Message:\n{job_details.message}"
                    )

                return status

            time.sleep(self.poll_interval)

    def _terminate(self, context: OpOrAssetExecutionContext, job_id: str) -> None:
        context.log.info(f"[pipes] Terminating RayJob {job_id} ...")
        self.client.stop_job(job_id)
        context.log.info(f"[pipes] Ray job {job_id} terminated.")



from dagster import Config, DagsterRunStatus, OpExecutionContext, RunsFilter, op
from pydantic import Field

from dagster_ray._base.constants import DEFAULT_DEPLOYMENT_NAME
from dagster_ray.kuberay.resources import RayClusterClientResource


class DeleteKubeRayClustersConfig(Config):
    namespace: str = "ray"
    cluster_names: list[str] = Field(default_factory=list, description="List of RayCluster names to delete")


@op(description="Deletes RayCluster resources from Kubernetes", name="delete_kuberay_clusters")
def delete_kuberay_clusters_op(
    context: OpExecutionContext,
    config: DeleteKubeRayClustersConfig,
    kuberay_client: RayClusterClientResource,
) -> None:
    for cluster_name in config.cluster_names:
        try:
            if kuberay_client.client.get(name=cluster_name, namespace=config.namespace).get("items"):
                kuberay_client.client.delete(name=cluster_name, namespace=config.namespace)
                context.log.info(f"RayCluster {config.namespace}/{cluster_name} deleted!")
            else:
                context.log.warning(f"RayCluster {config.namespace}/{cluster_name} doesn't exist")
        except Exception as e:  # noqa
            context.log.exception(f"Couldn't delete RayCluster {config.namespace}/{cluster_name}")


class CleanupKuberayClustersConfig(Config):
    namespace: str = "kuberay"
    label_selector: str = Field(
        default=f"dagster.io/deployment={DEFAULT_DEPLOYMENT_NAME}", description="Label selector to filter RayClusters"
    )


@op(
    description="Deletes RayCluster resources which do not correspond to any active Dagster Runs in this deployment from Kubernetes",
    name="cleanup_kuberay_clusters",
)
def cleanup_kuberay_clusters_op(
    context: OpExecutionContext,
    config: CleanupKuberayClustersConfig,
    kuberay_client: RayClusterClientResource,
) -> None:
    current_runs = context.instance.get_runs(
        filters=RunsFilter(
            statuses=[
                DagsterRunStatus.STARTED,
                DagsterRunStatus.QUEUED,
                DagsterRunStatus.CANCELING,
            ]
        )
    )

    clusters = kuberay_client.client.list(
        namespace=config.namespace,
        label_selector=config.label_selector,
    )["items"]

    # filter clusters by current runs using dagster.io/run_id label

    old_cluster_names = [
        cluster["metadata"]["name"]
        for cluster in clusters
        if not any(run.run_id == cluster["metadata"]["labels"]["dagster.io/run_id"] for run in current_runs)
    ]

    for cluster_name in old_cluster_names:
        try:
            kuberay_client.client.delete(name=cluster_name, namespace=config.namespace)
            context.log.info(f"RayCluster {config.namespace}/{cluster_name} deleted!")
        except:  # noqa
            context.log.exception(f"Couldn't delete RayCluster {config.namespace}/{cluster_name}")



from dagster import job

from dagster_ray.kuberay.ops import cleanup_kuberay_clusters_op, delete_kuberay_clusters_op


@job(description="Deletes RayCluster resources from Kubernetes", name="delete_kuberay_rayclusters")
def delete_kuberay_clusters():
    delete_kuberay_clusters_op()


@job(
    description="Deletes RayCluster resources which do not correspond to any active Dagster Runs in this deployment from Kubernetes",
    name="cleanup_kuberay_rayclusters",
)
def cleanup_kuberay_clusters():
    cleanup_kuberay_clusters_op()



from dagster_ray._base.resources import BaseRayResource
from dagster_ray.executor import ray_executor
from dagster_ray.io_manager import RayIOManager
from dagster_ray.pipes import PipesRayJobClient, PipesRayJobMessageReader
from dagster_ray.resources import LocalRay
from dagster_ray.run_launcher import RayRunLauncher

RayResource = BaseRayResource


__all__ = [
    "RayResource",
    "RayRunLauncher",
    "RayIOManager",
    "ray_executor",
    "PipesRayJobMessageReader",
    "PipesRayJobClient",
    "LocalRay",
]



import shutil
import subprocess
import sys
from pathlib import Path

import pytest

EXAMPLES_DIR = Path(__file__).parent.parent.parent / "examples"
LOCAL_EXAMPLES_DIR = EXAMPLES_DIR / "local"


RUN_LAUNCHER_EXAMPLE_DIR = LOCAL_EXAMPLES_DIR / "run_launcher"
EXECUTOR_EXAMPLE_DIR = LOCAL_EXAMPLES_DIR / "executor"
RUN_LAUNCHER_AND_EXECUTOR_EXAMPLE_DIR = LOCAL_EXAMPLES_DIR / "run_launcher_and_executor"


@pytest.mark.parametrize(
    "example_dir", [RUN_LAUNCHER_EXAMPLE_DIR, EXECUTOR_EXAMPLE_DIR, RUN_LAUNCHER_AND_EXECUTOR_EXAMPLE_DIR]
)
def test_ray_run_launcher(local_ray_address: str, example_dir: Path, tmp_path_factory):
    dagster_home = tmp_path_factory.mktemp("dagster_home")

    # copy dagter.yaml from example_dir to dagster_home

    shutil.copy(example_dir / "dagster.yaml", dagster_home / "dagster.yaml")

    # TODO: rewrite this test in Dagster's Python API instead of calling the CLI
    # the CLI command doesn't actually use the RunLauncher!

    subprocess.run(
        f"""cd {example_dir} && {sys.executable} -m dagster job execute -f {example_dir / "definitions.py"} -j my_job
        """,
        check=True,
        shell=True,
        env={
            "DAGSTER_HOME": str(dagster_home),
        },
    )



from __future__ import annotations

import uuid
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Union, cast

from dagster import AssetExecutionContext, ConfigurableResource, InitResourceContext, OpExecutionContext
from pydantic import Field, PrivateAttr

# yes, `python-client` is actually the KubeRay package name
# https://github.com/ray-project/kuberay/issues/2078
from requests.exceptions import ConnectionError
from tenacity import retry, retry_if_exception_type, stop_after_delay
from typing_extensions import TypeAlias

from dagster_ray._base.utils import get_dagster_tags
from dagster_ray.config import RayDataExecutionOptions
from dagster_ray.utils import _process_dagster_env_vars

if TYPE_CHECKING:
    from ray._private.worker import BaseContext as RayBaseContext  # noqa

OpOrAssetExecutionContext: TypeAlias = Union[OpExecutionContext, AssetExecutionContext]


class BaseRayResource(ConfigurableResource, ABC):
    """
    Base class for Ray Resources.
    Defines the common interface and some utility methods.
    """

    ray_init_options: dict[str, Any] = Field(
        default_factory=dict,
        description="Additional keyword arguments to pass to `ray.init()` call, such as `runtime_env`, `num_cpus`, etc. Dagster's `EnvVar` is supported. More details in [Ray docs](https://docs.ray.io/en/latest/ray-core/api/doc/ray.init.html).",
    )
    data_execution_options: RayDataExecutionOptions = Field(default_factory=RayDataExecutionOptions)
    redis_port: int = Field(
        default=10001, description="Redis port for connection. Make sure to match with the actual available port."
    )
    dashboard_port: int = Field(
        default=8265, description="Dashboard port for connection. Make sure to match with the actual available port."
    )
    enable_tracing: bool = Field(
        default=False,
        description="Enable tracing: inject `RAY_PROFILING=1` and `RAY_task_events_report_interval_ms=0` into the Ray cluster configuration. This allows using `ray.timeline()` to fetch recorded task events. Learn more: https://docs.ray.io/en/latest/ray-core/api/doc/ray.timeline.html#ray-timeline",
    )
    enable_actor_task_logging: bool = Field(
        default=False,
        description="Enable actor task logging: inject `RAY_ENABLE_RECORD_ACTOR_TASK_LOGGING=1` into the Ray cluster configuration.",
    )
    enable_debug_post_mortem: bool = Field(
        default=False,
        description="Enable post-mortem debugging: inject `RAY_DEBUG_POST_MORTEM=1` into the Ray cluster configuration. Learn more: https://docs.ray.io/en/latest/ray-observability/ray-distributed-debugger.html",
    )

    _context: RayBaseContext | None = PrivateAttr()

    def setup_for_execution(self, context: InitResourceContext) -> None:
        raise NotImplementedError(
            "This is an abstract resource, it's not meant to be provided directly. "
            "Use a backend-specific resource instead."
        )

    @property
    def context(self) -> RayBaseContext:
        assert self._context is not None, "RayClusterResource not initialized"
        return self._context

    @property
    @abstractmethod
    def host(self) -> str:
        raise NotImplementedError()

    @property
    def ray_address(self) -> str:
        return f"ray://{self.host}:{self.redis_port}"

    @property
    def dashboard_url(self) -> str:
        return f"http://{self.host}:{self.dashboard_port}"

    @property
    def runtime_job_id(self) -> str:
        """
        Returns the Ray Job ID for the current job which was created with `ray.init()`.
        :return:
        """
        import ray

        return ray.get_runtime_context().get_job_id()

    @retry(stop=stop_after_delay(120), retry=retry_if_exception_type(ConnectionError), reraise=True)
    def init_ray(self, context: OpOrAssetExecutionContext | InitResourceContext) -> RayBaseContext:
        assert context.log is not None

        import ray

        init_options = _process_dagster_env_vars(self.ray_init_options.copy())

        # cleanup None values from runtime_env.env_vars since Ray doesn't like them

        if "runtime_env" in init_options and "env_vars" in init_options["runtime_env"]:
            init_options["runtime_env"]["env_vars"] = {
                k: v for k, v in init_options["runtime_env"]["env_vars"].items() if v is not None
            }

        init_options["ignore_reinit_error"] = init_options.get("ignore_reinit_error", True)

        self.data_execution_options.apply()

        self._context = ray.init(
            address=self.ray_address,
            **init_options,
        )
        self.data_execution_options.apply()
        self.data_execution_options.apply_remote()
        context.log.info("Initialized Ray!")
        return cast("RayBaseContext", self._context)

    def get_dagster_tags(self, context: InitResourceContext) -> dict[str, str]:
        tags = get_dagster_tags(context)
        return tags

    def _get_step_key(self, context: InitResourceContext) -> str:
        # just return a random string
        # since we want a fresh cluster every time
        return str(uuid.uuid4())

    def get_env_vars_to_inject(self) -> dict[str, str]:
        vars: dict[str, str] = {}
        if self.enable_debug_post_mortem:
            vars["RAY_DEBUG_POST_MORTEM"] = "1"
        if self.enable_tracing:
            vars["RAY_PROFILING"] = "1"
            vars["RAY_task_events_report_interval_ms"] = "0"
        if self.enable_actor_task_logging:
            vars["RAY_ENABLE_RECORD_ACTOR_TASK_LOGGING"] = "1"
        return vars



import sys

import pytest
import ray  # noqa: TID253
from dagster import AssetExecutionContext, DagsterEventType, EventRecordsFilter, asset, materialize
from dagster._core.definitions.data_version import (
    DATA_VERSION_IS_USER_PROVIDED_TAG,
    DATA_VERSION_TAG,
)
from dagster._core.instance_for_test import instance_for_test
from pytest_kubernetes.providers import AClusterManager
from ray.job_submission import JobSubmissionClient  # noqa: TID253

from dagster_ray import PipesRayJobClient
from dagster_ray.kuberay.client import RayJobClient
from dagster_ray.kuberay.pipes import PipesKubeRayJobClient

ENTRYPOINT = "python /src/tests/scripts/remote_job.py"

RAY_JOB = {
    "apiVersion": "ray.io/v1",
    "kind": "RayJob",
    "metadata": {
        "annotations": {},
        "namespace": "ray",
    },
    "spec": {
        "activeDeadlineSeconds": 10800,
        "entrypoint": ENTRYPOINT,
        "entrypointNumCpus": 0.1,
        "rayClusterSpec": {
            "autoscalerOptions": {
                "imagePullPolicy": "IfNotPresent",
                "securityContext": {"runAsUser": 0},
            },
            "enableInTreeAutoscaling": False,
            "headGroupSpec": {
                "rayStartParams": {"dashboard-host": "0.0.0.0", "num-cpus": "1", "num-gpus": "0"},
                "serviceType": "ClusterIP",
                "template": {
                    "spec": {
                        "affinity": {},
                        "containers": [
                            {
                                "env": [],
                                "imagePullPolicy": "IfNotPresent",
                                "name": "ray-head",
                                "securityContext": {"runAsUser": 0},
                            }
                        ],
                        # "serviceAccountName": "ray",
                        "volumes": [],
                    },
                },
            },
            "rayVersion": ray.__version__,
            "workerGroupSpecs": [],
        },
        "shutdownAfterJobFinishes": False,  # helpful for debugging tests1
        "submissionMode": "K8sJobMode",
        # "ttlSecondsAfterFinished": 300,  # doesn't work together with shutdownAfterJobFinishes starting from KubeRay 1.4.0
    },
}


@pytest.fixture(scope="session")
def pipes_kube_rayjob_client(k8s_with_kuberay: AClusterManager):
    return PipesKubeRayJobClient(
        client=RayJobClient(
            config_file=str(k8s_with_kuberay.kubeconfig),
        ),
        port_forward=True,
    )


def test_rayjob_pipes(pipes_kube_rayjob_client: PipesKubeRayJobClient, dagster_ray_image: str, capsys):
    @asset
    def my_asset(context: AssetExecutionContext, pipes_kube_rayjob_client: PipesKubeRayJobClient):
        result = pipes_kube_rayjob_client.run(
            context=context,
            ray_job=RAY_JOB,
            extras={"foo": "bar"},
        ).get_materialize_result()

        return result

    with instance_for_test() as instance:
        result = materialize(
            [my_asset],
            resources={"pipes_kube_rayjob_client": pipes_kube_rayjob_client},
            instance=instance,
            tags={"dagster/image": dagster_ray_image},
        )

        captured = capsys.readouterr()

        print(captured.out)
        print(captured.err, file=sys.stderr)

        mat_evts = result.get_asset_materialization_events()

        mat = instance.get_latest_materialization_event(my_asset.key)
        instance.get_event_records(event_records_filter=EventRecordsFilter(event_type=DagsterEventType.LOGS_CAPTURED))

        assert len(mat_evts) == 1

        assert result.success
        assert mat
        assert mat and mat.asset_materialization
        assert mat.asset_materialization.metadata["some_metric"].value == 0
        assert mat.asset_materialization.tags
        assert mat.asset_materialization.tags[DATA_VERSION_TAG] == "alpha"
        assert mat.asset_materialization.tags[DATA_VERSION_IS_USER_PROVIDED_TAG]

        assert "Hello from stdout!" in captured.out
        assert "Hello from stderr!" in captured.out
        assert "Hello from Ray Pipes!" in captured.err


@pytest.fixture(scope="session")
def pipes_ray_job_client(k8s_with_raycluster: tuple[dict[str, str], AClusterManager]):
    hosts, k8s = k8s_with_raycluster
    return PipesRayJobClient(
        client=JobSubmissionClient(
            address=hosts["dashboard"],
        )
    )


def test_ray_job_pipes(pipes_ray_job_client: PipesRayJobClient, capsys):
    @asset
    def my_asset(context: AssetExecutionContext, pipes_ray_job_client: PipesRayJobClient):
        result = pipes_ray_job_client.run(
            context=context,
            submit_job_params={"entrypoint": ENTRYPOINT, "entrypoint_num_cpus": 0.1},
            extras={"foo": "bar"},
        ).get_materialize_result()

        return result

    with instance_for_test() as instance:
        result = materialize(
            [my_asset],
            resources={"pipes_ray_job_client": pipes_ray_job_client},
            instance=instance,
        )

        captured = capsys.readouterr()

        print(captured.out)
        print(captured.err, file=sys.stderr)

        mat_evts = result.get_asset_materialization_events()

        mat = instance.get_latest_materialization_event(my_asset.key)
        instance.get_event_records(event_records_filter=EventRecordsFilter(event_type=DagsterEventType.LOGS_CAPTURED))

        assert len(mat_evts) == 1

        assert result.success
        assert mat
        assert mat and mat.asset_materialization
        assert mat.asset_materialization.metadata["some_metric"].value == 0
        assert mat.asset_materialization.tags
        assert mat.asset_materialization.tags[DATA_VERSION_TAG] == "alpha"
        assert mat.asset_materialization.tags[DATA_VERSION_IS_USER_PROVIDED_TAG]

        assert "Hello from stdout!" in captured.out
        assert "Hello from stderr!" in captured.out
        assert "Hello from Ray Pipes!" in captured.err



from __future__ import annotations

import os
from collections.abc import Mapping, Sequence
from typing import Any

import dagster as dg
from dagster._config.field_utils import IntEnvVar

# yes, `python-client` is actually the KubeRay package name
# https://github.com/ray-project/kuberay/issues/2078


def resolve_env_vars_list(env_vars: list[str] | None) -> dict[str, str]:
    res = {}

    if env_vars is not None:
        for env_var in env_vars:
            if "=" in env_var:
                var, value = env_var.split("=", 1)
                res[var] = value
            else:
                if value := os.getenv(env_var):
                    res[env_var] = value

    return res


def _process_dagster_env_vars(config: Any) -> Any:
    # If it's already one of the EnvVar types, return its value
    if isinstance(config, (dg.EnvVar, IntEnvVar)):
        return config.get_value()

    # If it's exactly {"env": "FOO"}, fetch the value
    if isinstance(config, Mapping) and len(config) == 1 and "env" in config:
        return dg.EnvVar(config["env"]).get_value()

    # If it's a dictionary, recurse into its values
    if isinstance(config, Mapping):
        return {k: _process_dagster_env_vars(v) for k, v in config.items()}

    # If it's a list/tuple, recurse into each element
    if isinstance(config, Sequence) and not isinstance(config, str):
        return [_process_dagster_env_vars(v) for v in config]

    # Otherwise, just return as-is
    return config
